/*
** AUTHORS:
** Rama Alebouyeh (rama[at]cs.ucsb.edu)
** Matthew S Allen (msa[at]cs.ucsb.edu)
*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <openssl/evp.h>
#include "key.h"
#include "log.h"

void key_print(Key k){
  int i;
  for(i=4; i>=0; i--)
    printf("%08x ",(unsigned int)k.t[i]);
  printf("\n");
}

void key_to_str(char *str, Key k){
  memset(str,KEY_SIZE/BASE_B,0);
  sprintf(str,"%08x%08x%08x%08x%08x",(unsigned int)k.t[4],(unsigned int)k.t[3],
	  (unsigned int)k.t[2],(unsigned int)k.t[1],(unsigned int)k.t[0]);
}

void str_to_key(char *str, Key *k){
  
  int i,len;
  char key_str[KEY_SIZE/BASE_B+1];
  
  for(i=0;i<KEY_SIZE/BASE_B;i++)
    key_str[i]='0';
  key_str[KEY_SIZE/BASE_B]='\0';
   
  
  len = strlen(str);
  if(len==0){
    fprintf(stderr,"str_to_key: Warning:Empty input string\n");
  }
  else if(len> KEY_SIZE/BASE_B){
    strncpy(key_str,str,KEY_SIZE/BASE_B);
    key_str[KEY_SIZE/BASE_B]='\0';
  }

  else if(len <= KEY_SIZE/BASE_B){
    for(i=0;i<len;i++)
      key_str[i+(KEY_SIZE/BASE_B)-len]=str[i];
  }
      
  for(i=0; i <5; i++)
    sscanf(key_str+(i*8*sizeof(char)),"%08x",&(k->t[(4-i)]));
  
  key_to_str(k->keystr,*k);
}


void key_assign(Key *k1, Key k2){
  int i;
  for(i=0;i<5;i++)
    k1->t[i]=k2.t[i];
  
  key_to_str(k1->keystr,*k1);
}

void key_assign_ui(Key *k,unsigned long ul){
  int i;
  for(i=0;i<5;i++)
    k->t[i]=0;
  k->t[0]=ul;
  key_to_str(k->keystr,*k); 
}

int key_equal(Key k1,Key k2){
  int i;
  for(i=0; i<5; i++)
    if(k1.t[i]!=k2.t[i])
      return(0);
  return(1); 
}

int key_equal_ui(Key k,unsigned long ul){
  int i;
  if(k.t[0]!=ul)
    return(0);
  for(i=1;i<5;i++)
    if(k.t[i]!=0)
      return(0);
  return(1);
}

int key_comp(Key k1, Key k2){

  int i;
  
  if(key_equal(k1,k2))
    return (0);
  
  for(i=4;i>=0;i--){
    if(k1.t[i] > k2.t[i])
      return (1);
    else if(k1.t[i] < k2.t[i])
      return (-1);
  }
  return(0);
}

void key_add(Key *result, Key op1, Key op2){

  double tmp,a,b;
  int i;
  a=b=tmp=0;

  for(i=0;i<5;i++){
    a= op1.t[i];
    b= op2.t[i];
    tmp += a + b;
    
    if ( tmp > ULONG_MAX){
      result->t[i] = (unsigned long)tmp;
      tmp = 1;
    }
    else {
      result->t[i] =(unsigned long) tmp; 
      tmp =0;
    }
  }
  key_to_str(result->keystr,*result);
}

void key_sub(void * logs, Key *result, Key op1, Key op2){
  int i;
  double tmp,a,b,carry;

  carry = 0;
  
  if(key_comp(op1,op2)<0){
    log_message(logs, LOG_ERROR,"key_sub: Operation is not allowed %s < %s \n", op1.keystr,op2.keystr);
    return;
  }
 
  for(i=0;i<5;i++){
    a = op1.t[i] - carry ;
    b = op2.t[i];

    if(b <= a){
      tmp = a-b;
      carry = 0;
    }
    else {
      a=a+ULONG_MAX+1;
      tmp = a-b;
      carry = 1;
    }
    result->t[i] = (unsigned long) tmp; 
   }

  key_to_str(result->keystr,*result);
}


char * sha1_keygen(char * key, char * digest, int size )
{
  EVP_MD_CTX mdctx;
  const EVP_MD *md;
  unsigned char *md_value; 
  int md_len, i;
  
  md_value = (unsigned char *)malloc(EVP_MAX_MD_SIZE);
  digest = (char *) malloc (KEY_SIZE/BASE_B);

  OpenSSL_add_all_digests();

  md = EVP_get_digestbyname("sha1");

  EVP_MD_CTX_init(&mdctx);
  EVP_DigestInit_ex(&mdctx, md, NULL);
  EVP_DigestUpdate(&mdctx, key, strlen(key));
  EVP_DigestFinal_ex(&mdctx, md_value, &md_len);
  EVP_MD_CTX_cleanup(&mdctx);

  
  char * tmp= digest;
  for(i = 0; i < md_len; i++){
    sprintf(tmp,"%02x", md_value[i]);
    tmp=tmp+2;
  }

  tmp='\0';
  digest[size]='\0';
  
  return(digest);
}

void key_makehash(void *logs, Key *hashed, char *s) {
  
  char * digest;
  int i;

  digest=sha1_keygen(s,digest,KEY_SIZE/BASE_B);

  for(i=0; i <5; i++)
    sscanf(digest+(i*8*sizeof(char)),"%08x",&hashed->t[(4-i)]);

  key_to_str(hashed->keystr,*hashed);
  log_message(logs, LOG_KEYDEBUG,"key_makehash: HASH( %s ) = [%s]\n",s,hashed->keystr);
}

void key_init(){
  int i;
  for(i=0;i<5;i++){
    Key_Max.t[i]=ULONG_MAX;
    Key_Half.t[i]=ULONG_MAX;
  }
  Key_Half.t[4]= Key_Half.t[4]/2;
  
  key_to_str(Key_Max.keystr,Key_Max);
  key_to_str(Key_Half.keystr,Key_Half); 
  
}

void key_distance(void * logs, Key *diff, Key k1, Key k2) {

  int comp=0;
  
  comp=key_comp(k1,k2);
  
  /* k1 > k2 */
  if(comp > 0)
    key_sub (logs,diff, k1, k2);
  else
    key_sub (logs,diff, k2, k1);

  comp=0;
  comp = key_comp (*diff,Key_Half);
  
  /* diff > Key_Half */
  if(comp > 0) 
    key_sub (logs, diff, Key_Max,*diff);
 

  key_to_str(diff->keystr,*diff);  
  log_message(logs, LOG_KEYDEBUG, "key_distance: %s  and %s is %s\n", k1.keystr, k2.keystr, diff->keystr);
}

int key_between(void *logs, Key test, Key left, Key right) {
  int comp;
  Key t,l,r;
  Key tmp;
  
  comp= key_comp(left,right);
  if(comp==0)
    return(0);

  key_assign_ui(&l,0);

  if(key_comp(left,test)<0){
    key_sub(logs, &t,test, left);
  } else {
    key_sub(logs,&tmp, Key_Max, left);
    key_add(&t, test, tmp);
  }
  
  if(key_comp(left,right)<0){
    key_sub(logs, &r, right, left);
  } else {
    key_sub(logs,&tmp, Key_Max, left);
    key_add(&r, right, tmp);
  }
    
  log_message(logs, LOG_KEYDEBUG, "key_between: %s %s between %s and %s\n", test.keystr, (key_comp(t,r)<0)?("IS"):("IS NOT"), left.keystr, right.keystr);

  if(key_comp(t,r)<0){
    return(1);
  }
  return(0);
}
void key_midpoint(void * logs, Key *mid, Key key){
  
  if(key_comp(key,Key_Half)<0)
    key_add(mid,key,Key_Half);
  else
    key_sub(logs,mid,key,Key_Half);
  
  key_to_str(mid->keystr,*mid);
}

int key_index(void *logs, Key mykey, Key k){

  int max_len,i;
  char mystr[KEY_SIZE/BASE_B];
  char kstr[KEY_SIZE/BASE_B];
  max_len = KEY_SIZE/BASE_B;
  key_to_str(mystr, mykey);
  key_to_str(kstr,k);
  
  for(i=0; (mystr[i]==kstr[i]) && (i<max_len); i++);
  if(i==max_len)
    i=max_len-1;

  log_message(logs, LOG_KEYDEBUG, "key_index:%d\n me:%s\n lookup_key:%s \n",i,mykey.keystr,k.keystr);
 
  return(i);
}

